/* 
 *  QuantDump Copyright (C) 2005-2006, Donald A. Graft
 *  Dumps quant matrices from an elementary or program stream.
 *  usage: quantdump infile [outfile]
 *
 *  QuantDump is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  This file is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 */

#include <windows.h>
#include <stdio.h> 
#include <io.h> 
#include <fcntl.h>

#define ELEMENTARY_STREAM 0
#define MPEG1_PROGRAM_STREAM 1
#define MPEG2_PROGRAM_STREAM 2

// Defines for the start code detection state machine.
#define NEED_FIRST_0  0
#define NEED_SECOND_0 1
#define NEED_1        2

static void determine_stream_type(void);
static void video_parser(void);
static void pack_parser(void);
static unsigned char get_byte(void);

static int stream_type;
static int file;
static FILE *wfp;
static int state, found;
// Should hold maximum size PES packet.
static unsigned char buffer[256000];
static int buffer_length, buffer_ndx;
static int EOF_reached;

unsigned char intra_quant_matrix[64];
unsigned char intra_quant_matrix_old[64];
unsigned char non_intra_quant_matrix[64];
unsigned char non_intra_quant_matrix_old[64];
unsigned char chroma_intra_quant_matrix[64];
unsigned char chroma_intra_quant_matrix_old[64];
unsigned char chroma_non_intra_quant_matrix[64];
unsigned char chroma_non_intra_quant_matrix_old[64];

/* default intra quantization matrix */
unsigned char default_intra_quantizer_matrix[64]
=
{
	8, 16, 19, 22, 26, 27, 29, 34,
	16, 16, 22, 24, 27, 29, 34, 37,
	19, 22, 26, 27, 29, 34, 34, 38,
	22, 22, 26, 27, 29, 34, 37, 40,
	22, 26, 27, 29, 32, 35, 40, 48,
	26, 27, 29, 32, 35, 40, 48, 58,
	26, 27, 29, 34, 38, 46, 56, 69,
	27, 29, 35, 38, 46, 56, 69, 83
};

/* zig-zag scan pattern */
unsigned char scan[64]
=
{ // Zig-Zag scan pattern
	0,  1,  8, 16,  9,  2,  3, 10,
   17, 24, 32, 25, 18, 11,  4,  5,
   12, 19, 26, 33, 40, 48, 41, 34,
   27, 20, 13,  6,  7, 14, 21, 28,
   35, 42, 49, 56, 57, 50, 43, 36,
   29, 22, 15, 23, 30, 37, 44, 51,
   58, 59, 52, 45, 38, 31, 39, 46,
   53, 60, 61, 54, 47, 55, 62, 63
};

int initial_parse(char *input_file, char *output_file)
{
	// Open the input file.
	if (input_file[0] == 0 || (file = _open(input_file, _O_RDONLY | _O_BINARY | _O_SEQUENTIAL)) == -1)
	{
		printf("could not open the input file\n");
		return -1;
	}

	// Open the output file.
	if (!strcmp(output_file, "stdout"))
	{
		wfp = stdout;
	}
	else if ((wfp = fopen(output_file, "w")) == 0)
	{
		_close(file);
		printf("could not open the output file\n");
		return -1;
	}

	// Determine the stream type: ES or program.
	determine_stream_type();

	// Re-open the input file.
	_close(file);
	file = _open(input_file, _O_RDONLY | _O_BINARY | _O_SEQUENTIAL);
	if (file == -1)
	{
		return -1;
	}

	// Determine MPEG type and find location of first video sequence header.
	EOF_reached = 0;
	video_parser();
	_close(file);
	fclose(wfp);
	if (EOF_reached)
	{
		return -1;
	}
	return 0;
}

static unsigned char get_byte(void)
{
	unsigned char val;

	if (stream_type != ELEMENTARY_STREAM)
	{
		if (buffer_ndx >= buffer_length)
		{
			// Need more data. Fill buffer with the next
			// packet's worth of elementary data.
			pack_parser();
		}
		if (EOF_reached == 1) 
		{
			return 0;
		}
		return buffer[buffer_ndx++];
	}
	else
	{
		// No program stream parsing needed. Just read from
		// the input file.
		if (_read(file, &val, 1) != 1)
		{
			EOF_reached = 1;
			return 0;
		}
		return val;
	}
}

static void video_parser()
{
	int i;
	unsigned char val;
	int sequence_header_active = 0;
	int num_fields = 0;

	buffer_length = buffer_ndx = 0;

	// Inits.
	state = NEED_FIRST_0;
	found = 0;

	// Default quant matrices.
	fprintf(wfp, "Default Luma and Chroma Intra Quant Matrix:\n");
	for (i=0; i<64; i++)
	{
		fprintf(wfp, "%d ", default_intra_quantizer_matrix[i]);
		if ((i % 8) == 7)
			fprintf(wfp, "\n");
	}
	fprintf(wfp, "\n");
	fprintf(wfp, "Default Luma and Chroma Non-Intra Quant Matrix:\n");
	for (i=0; i<64; i++)
	{
		fprintf(wfp, "%d ", 16);
		if ((i % 8) == 7)
			fprintf(wfp, "\n");
	}
	fprintf(wfp, "\n");

	while(1)
	{
		// Parse for start codes.
		val = get_byte();
		if (EOF_reached) return;
		switch (state)
		{
		case NEED_FIRST_0:
			if (val == 0)
				state = NEED_SECOND_0;
			break;
		case NEED_SECOND_0:
			if (val == 0)
				state = NEED_1;
			else
				state = NEED_FIRST_0;
			break;
		case NEED_1:
			if (val == 1)
			{
				found = 1;
				state = NEED_FIRST_0;
			}
			else if (val != 0)
				state = NEED_FIRST_0;
			break;
		}
		if (found == 1)
		{
			// Found a start code.
			found = 0;
			val = get_byte();
			if (EOF_reached) return;

			if (val == 0xB3)
			{
				// Found a sequence header.
				// Get quant matrix.
				val = get_byte();
				if (EOF_reached) return;
				val = get_byte();
				if (EOF_reached) return;
				val = get_byte();
				if (EOF_reached) return;
				val = get_byte();
				if (EOF_reached) return;
				val = get_byte();
				if (EOF_reached) return;
				val = get_byte();
				if (EOF_reached) return;
				val = get_byte();
				if (EOF_reached) return;
				val = get_byte();
				if (EOF_reached) return;
				if (val & 0x02)
				{
					// Have an intra quant matrix.
					for (i = 0; i < 64; i++)
					{
						intra_quant_matrix[scan[i]] = (val & 0x01) << 7;
						val = get_byte();
						if (EOF_reached) return;
						intra_quant_matrix[scan[i]] |= (val & 0xfe) >> 1;
					}
					for (i = 0; i < 64; i++)
					{
						if (intra_quant_matrix[i] != intra_quant_matrix_old[i])
							break;
					}
					if (i < 64)
					{
						fprintf(wfp, "Luma and Chroma Intra Quant Matrix at encoded frame %d:\n", num_fields / 2);
						for (i = 0; i < 64; i++)
						{
							fprintf(wfp, "%d ", intra_quant_matrix[i]);
							if ((i % 8) == 7)
								fprintf(wfp, "\n");
						}
						fprintf(wfp, "\n");
					}
					memcpy(intra_quant_matrix_old, intra_quant_matrix, sizeof(intra_quant_matrix));
				}
				if (val & 0x01)
				{
					// Have a non intra quant matrix.
					for (i = 0; i < 64; i++)
					{
						val = get_byte();
						if (EOF_reached) return;
						non_intra_quant_matrix[scan[i]] = val;
					}
					for (i = 0; i < 64; i++)
					{
						if (non_intra_quant_matrix[i] != non_intra_quant_matrix_old[i])
							break;
					}
					if (i < 64)
					{
						fprintf(wfp, "Luma and Chroma Non-Intra Quant Matrix at encoded frame %d:\n", num_fields / 2);
						for (i = 0; i < 64; i++)
						{
							fprintf(wfp, "%d ", non_intra_quant_matrix[i]);
							if ((i % 8) == 7)
								fprintf(wfp, "\n");
						}
						fprintf(wfp, "\n");
					}
					memcpy(non_intra_quant_matrix_old, non_intra_quant_matrix, sizeof(non_intra_quant_matrix));
				}
			}
			else if (val == 0xB5)
			{
				val = get_byte();
				if (EOF_reached) return;
				if ((val & 0xf0) == 0x80)
				{
					// Picture coding extension.
					val = get_byte();
					if (EOF_reached) return;
					val = get_byte();
					if (EOF_reached) return;
					if ((val & 3) == 3)
						// Count two fields.
						num_fields += 2;
					else
						// Count one field.
						num_fields += 1;
				}
				else if ((val & 0xf0) == 0x30)
				{
					// Quant matrix extension.
					if (val & 0x08)
					{
						// Have a luma intra quant matrix.
						for (i = 0; i < 64; i++)
						{
							intra_quant_matrix[scan[i]] = (val & 0x07) << 5;
							val = get_byte();
							if (EOF_reached) return;
							intra_quant_matrix[scan[i]] |= (val & 0xf8) >> 3;
						}
						for (i = 0; i < 64; i++)
						{
							if (intra_quant_matrix[i] != intra_quant_matrix_old[i])
								break;
						}
						if (i < 64)
						{
							fprintf(wfp, "Luma Intra Quant Matrix at encoded frame %d:\n", num_fields / 2);
							for (i = 0; i < 64; i++)
							{
								fprintf(wfp, "%d ", intra_quant_matrix[i]);
								if ((i % 8) == 7)
									fprintf(wfp, "\n");
							}
							fprintf(wfp, "\n");
						}
						memcpy(intra_quant_matrix_old, intra_quant_matrix, sizeof(intra_quant_matrix));
					}
					if (val & 0x04)
					{
						// Have a luma non intra quant matrix.
						for (i = 0; i < 64; i++)
						{
							non_intra_quant_matrix[scan[i]] = (val & 0x03) << 6;
							val = get_byte();
							if (EOF_reached) return;
							non_intra_quant_matrix[scan[i]] |= (val & 0xfc) >> 2;
						}
						for (i = 0; i < 64; i++)
						{
							if (non_intra_quant_matrix[i] != non_intra_quant_matrix_old[i])
								break;
						}
						if (i < 64)
						{
							fprintf(wfp, "Luma Non-Intra Quant Matrix at encoded frame %d:\n", num_fields / 2);
							for (i = 0; i < 64; i++)
							{
								fprintf(wfp, "%d ", non_intra_quant_matrix[i]);
								if ((i % 8) == 7)
									fprintf(wfp, "\n");
							}
							fprintf(wfp, "\n");
						}
						memcpy(non_intra_quant_matrix_old, non_intra_quant_matrix, sizeof(non_intra_quant_matrix));
					}
					if (val & 0x02)
					{
						// Have a chroma intra quant matrix.
						for (i = 0; i < 64; i++)
						{
							chroma_intra_quant_matrix[scan[i]] = (val & 0x01) << 7;
							val = get_byte();
							if (EOF_reached) return;
							chroma_intra_quant_matrix[scan[i]] |= (val & 0xfe) >> 1;
						}
						for (i = 0; i < 64; i++)
						{
							if (chroma_intra_quant_matrix[i] != chroma_intra_quant_matrix_old[i])
								break;
						}
						if (i < 64)
						{
							fprintf(wfp, "Chroma Intra Quant Matrix at encoded frame %d:\n", num_fields / 2);
							for (i = 0; i < 64; i++)
							{
								fprintf(wfp, "%d ", chroma_intra_quant_matrix[i]);
								if ((i % 8) == 7)
									fprintf(wfp, "\n");
							}
							fprintf(wfp, "\n");
						}
						memcpy(chroma_intra_quant_matrix_old, chroma_intra_quant_matrix, sizeof(chroma_intra_quant_matrix));
					}
					if (val & 0x01)
					{
						// Have a chroma non intra quant matrix.
						for (i = 0; i < 64; i++)
						{
							val = get_byte();
							if (EOF_reached) return;
							chroma_non_intra_quant_matrix[scan[i]] = val;
						}
						for (i = 0; i < 64; i++)
						{
							if (chroma_non_intra_quant_matrix[i] != chroma_non_intra_quant_matrix_old[i])
								break;
						}
						if (i < 64)
						{
							fprintf(wfp, "Chroma Non-Intra Quant Matrix at encoded frame %d:\n", num_fields / 2);
							for (i = 0; i < 64; i++)
							{
								fprintf(wfp, "%d ", chroma_non_intra_quant_matrix[i]);
								if ((i % 8) == 7)
									fprintf(wfp, "\n");
							}
							fprintf(wfp, "\n");
						}
						memcpy(chroma_non_intra_quant_matrix_old, chroma_non_intra_quant_matrix, sizeof(chroma_non_intra_quant_matrix));
					}
				}
			}
		}
	}
	return;
}

static void pack_parser(void)
{
	unsigned char val, val2;
	int state, found = 0;
	int length;
	unsigned short pes_packet_length, system_header_length;
	unsigned char pes_packet_header_length, pack_stuffing_length;
	unsigned char code;

	// Look for start codes.
	state = NEED_FIRST_0;
	while (1)
	{
		if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
		switch (state)
		{
		case NEED_FIRST_0:
			if (val == 0)
				state = NEED_SECOND_0;
			break;
		case NEED_SECOND_0:
			if (val == 0)
				state = NEED_1;
			else
				state = NEED_FIRST_0;
			break;
		case NEED_1:
			if (val == 1)
			{
				found = 1;
				state = NEED_FIRST_0;
			}
			else if (val != 0)
				state = NEED_FIRST_0;
			break;
		}
		if (found == 1)
		{
			// Found a start code.
			found = 0;
			// Get the start code.
			if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
//			dprintf("DGIndex: code = %x\n", val);
			if (val == 0xba)
			{
				// Pack header. Skip it.
				if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
				if (stream_type == MPEG1_PROGRAM_STREAM)
				{
					// MPEG1 program stream.
					if (_read(file, &buffer, 7) != 7) { EOF_reached = 1; return; }
				}
				else
				{
					// MPEG2 program stream.
					if (_read(file, &buffer, 8) != 8) { EOF_reached = 1; return; }
					if (_read(file, &pack_stuffing_length, 1) != 1) { EOF_reached = 1; return; }
					pack_stuffing_length &= 0x03;
					if (_read(file, &buffer, pack_stuffing_length) != pack_stuffing_length) { EOF_reached = 1; return; }
				}
			}
			if (val == 0xbb)
			{
				// System header. Skip it.
				if (_read(file, &val, 1) != 1)  { EOF_reached = 1; return; }
				system_header_length = val << 8;
				if (_read(file, &val2, 1) != 1)  { EOF_reached = 1; return; }
				system_header_length |= val2;
				if (_read(file, &buffer, system_header_length) != system_header_length) { EOF_reached = 1; return; }
			}
			else if (val >= 0xE0 && val <= 0xEF)
			{
				// Video stream.
				// Packet length.
				if (_read(file, &val, 1) != 1)  { EOF_reached = 1; return; }
				pes_packet_length = val << 8;
				if (_read(file, &val2, 1) != 1)  { EOF_reached = 1; return; }
				pes_packet_length |= val2;
				if (stream_type == MPEG1_PROGRAM_STREAM)
				{
					// MPEG1 program stream.
					pes_packet_header_length = 0;
					// Stuffing bytes.
					do 
					{
						if (_read(file, &val, 1) != 1)  { EOF_reached = 1; return; }
						pes_packet_header_length += 1;
					} while (val == 0xff);
					if ((val & 0xc0) == 0x40)
					{
						// STD bytes.
						if (_read(file, &val, 1) != 1)  { EOF_reached = 1; return; }
						if (_read(file, &val, 1) != 1)  { EOF_reached = 1; return; }
						pes_packet_header_length += 2;
					}
					if ((val & 0xf0) == 0x20)
					{
						// PTS bytes.
						if (_read(file, &buffer, 4) != 4)  { EOF_reached = 1; return; }
						pes_packet_header_length += 4;
					}
					else if ((val & 0xf0) == 0x30)
					{
						// PTS/DTS bytes.
						if (_read(file, &buffer, 9) != 9)  { EOF_reached = 1; return; }
						pes_packet_header_length += 9;
					}
					// Send the video elementary data down the pipe to the video parser.
					buffer_length = pes_packet_length - pes_packet_header_length;
					if (buffer_length)
					{
						buffer_ndx = 0;
						if (_read(file, buffer, buffer_length) != buffer_length)  { EOF_reached = 1; return; }
						return;
					}
				}
				else
				{
					// MPEG2 program stream.
					// Flags.
					if (_read(file, &code, 1) != 1)  { EOF_reached = 1; return; }
					if ((code & 0xc0) == 0x80)
					{
						if (_read(file, &code, 1) != 1)  { EOF_reached = 1; return; }
						if (_read(file, &pes_packet_header_length, 1) != 1) { EOF_reached = 1; return; }
						// Skip the PES packet header.
						if (_read(file, buffer, pes_packet_header_length) != pes_packet_header_length)  { EOF_reached = 1; return; }
						// Send the video elementary data down the pipe to the video parser.
						buffer_length = pes_packet_length - pes_packet_header_length - 3;
						buffer_ndx = 0;
						if (_read(file, buffer, buffer_length) != buffer_length)  { EOF_reached = 1; return; }
						return;
					}
					else
					{
						// No video data here. Skip it.
						length = pes_packet_length - 1;
						if (_read(file, buffer, length) != length)  { EOF_reached = 1; return; }
					}
				}
			}
			else if (val > 0xbb)
			{
				// Not a stream that we are interested in. Skip it.
				if (_read(file, &val, 1) != 1)  { EOF_reached = 1; return; }
				pes_packet_length = val << 8;
				if (_read(file, &val2, 1) != 1)  { EOF_reached = 1; return; }
				pes_packet_length |= val2;
				length = pes_packet_length;
				if (_read(file, buffer, length) != length)  { EOF_reached = 1; return; }
			}
		}
	}
}

static void determine_stream_type(void)
{
	int i;
	unsigned char val;
	int state, found = 0;
	int num_sequence_header = 0;
	int num_picture_header = 0;

	// Start by assuming ES. Then look for a valid pack start. If one
	// is found declare a program stream.
	stream_type = ELEMENTARY_STREAM;

	// Look for start codes.
	state = NEED_FIRST_0;
	// Look for a pack start in the first part of the file.
	for (i = 0; i < 2500000; i++)
	{
		if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
		switch (state)
		{
		case NEED_FIRST_0:
			if (val == 0)
				state = NEED_SECOND_0;
			break;
		case NEED_SECOND_0:
			if (val == 0)
				state = NEED_1;
			else
				state = NEED_FIRST_0;
			break;
		case NEED_1:
			if (val == 1)
			{
				found = 1;
				state = NEED_FIRST_0;
			}
			else if (val != 0)
				state = NEED_FIRST_0;
			break;
		}
		if (found == 1)
		{
			// Found a start code.
			found = 0;
			// Get the start code.
			if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
			if (val == 0xba)
			{
				if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
				if ((val & 0xf0) == 0x20)
				{
					// Check all the marker bits just to be sure.
					if (!(val & 1)) continue;
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (!(val & 1)) continue;
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (!(val & 1)) continue;
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (!(val & 0x80)) continue;
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (!(val & 1)) continue;
					// MPEG1 program stream.
					stream_type = MPEG1_PROGRAM_STREAM;
					break;
				}
				else if ((val & 0xc0) == 0x40)
				{
					// Check all the marker bits just to be sure.
					if (!(val & 0x04)) continue;
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (!(val & 0x04)) continue;
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (!(val & 0x04)) continue;
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (!(val & 0x01)) continue;
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (_read(file, &val, 1) != 1) { EOF_reached = 1; return; }
					if (!(val & 0x03)) continue;
					// MPEG2 program stream.
					stream_type = MPEG2_PROGRAM_STREAM;
					break;
				}
			}
			else if (val == 0xb3)
			{
				// Sequence header.
				num_sequence_header++;
			}
			else if (val == 0x00)
			{
				// Picture header.
				num_picture_header++;
			}
			if (num_sequence_header >= 2 && num_picture_header >= 2)
			{
				// We're seeing a lot of elementary stream data but we haven't seen
				// a pack header yet. Declare ES.
				break;
			}
		}
	}
}

main(int argc, char **argv)
{
	if (argc < 2)
	{
		printf("QuantDump 1.1 by Donald A. Graft\n");
		printf("Usage: quantdump inputfile [outputfile]\n");
		printf("       where input file is an elementary or program stream\n");
		printf("       and outputfile will be written with the quant dump information.\n");
		return 0;
	}
	if (argc < 3)
		initial_parse(argv[1], "stdout");
	else
		initial_parse(argv[1], argv[2]);
	return 0;
}